import numpy as np
from scipy.linalg import expm


class fitcph:
    """
    Fits continuous-time phase-type distributions using the
    EM algorithm from p. 678 Bladt and Nielsen (2017).

    References:
        Bladt, M., & Nielsen, B. F. (2017). Matrix-Exponential Distributions in Applied Probability.
        Springer. https://doi.org/10.1007/978-1-4939-7049-0
    """

    def __init__(
        self,
        obs: np.array = None,
        initpi: np.array = None,
        initphgen: np.array = None,
        initexitrates: np.array = None,
        randominit: bool = True,
        seed: int = None,
        tolerance: float = 1e-6,
        itermax: int = 1000000,
        verbose: bool = False,
    ) -> None:
        """
        Initializes the continuous-time phase-type distribution fitter.

        Args:
            obs (array-like): Observed realizations of the phase-type distribution.
            initpi (ndarray): Initial distribution vector.
            initphgen (ndarray): Initial phase-type generator matrix.
            initexitrates (ndarray): Initial exit rate vector.
            randominit (bool, default=True): Whether to randomize initial parameters.
            seed (int): Random seed for reproducibility.
            tolerance (float, default=1e-6): Convergence tolerance for the EM algorithm.
            itermax (int, default=1000000): Maximum number of EM iterations.
            verbose (bool, default=False): Whether to print intermediate fitting information.
        """
        self.obs = obs  # observed realizations of the PH distribution
        self.initpi = initpi
        self.initphgen = initphgen
        self.initexitrates = initexitrates
        self.nphases = self.initphgen.shape[0]
        self.randominit = randominit
        self.seed = seed
        self.itermax = itermax
        self.tolerance = tolerance
        self.verbose = verbose
        self.__initialize()

    def fit(self) -> None:
        """
        Fits the continuous-time phase-type distribution using the EM algorithm.

        Args:
            None

        Returns:
            None
        """
        # fit the CPH distribution
        iter = 0
        eps = np.inf
        loglik0 = -np.inf
        while iter < self.itermax and eps > self.tolerance:
            self.__estep()
            self.__mstep()
            eps = self.loglikelihood - loglik0  # loglik is evaluated within the E-step
            loglik0 = self.loglikelihood
            iter += 1
            if self.verbose and iter % 25 == 0:
                if isinstance(eps, float):
                    printeps = eps
                else:
                    printeps = eps.item()
                print(
                    "iter =",
                    iter,
                    "  eps =",
                    printeps,
                    "  mean =",
                    self.getmean(),
                    "  var =",
                    self.getvar(),
                )
        self.__polish()
        self.__updatelikelihood()  # evaluate final loglik

    def getinitdist(self) -> np.array:
        """
        Returns the initial distribution vector.

        Args:
            None

        Returns:
            ndarray: The initial distribution.
        """
        return self.pi

    def getphasegen(self) -> np.array:
        """
        Returns the phase-type generator matrix.

        Args:
            None

        Returns:
            ndarray: The phase-type generator.
        """
        return self.phgen

    def getexitrates(self) -> np.array:
        """
        Returns the exit rate vector.

        Args:
            None

        Returns:
            ndarray: The exit rates.
        """
        return self.exitrates

    def getmean(self) -> float:
        """
        Returns the mean of the continuous-time phase-type distribution.

        Args:
            None

        Returns:
            float: The mean of the distribution.
        """
        return -np.sum(np.matmul(self.pi, np.linalg.inv(self.phgen)))

    def getvar(self) -> float:
        """
        Returns the variance of the continuous-time phase-type distribution.

        Args:
            None

        Returns:
            float: The variance of the distribution.
        """
        phinv = np.linalg.inv(self.phgen)
        return 2 * np.sum(
            np.matmul(self.pi, np.linalg.matrix_power(phinv, 2))
        ) - np.power(np.sum(np.matmul(self.pi, phinv)), 2)

    def getdensity(self, x: float) -> float:
        """
        Returns the distribution's density, f(x).

        Args:
            x (float): The distribution's density will be computed at time of x.

        Returns:
            float: The computed density.
        """
        return np.matmul(
            self.pi, np.matmul(expm(self.phgen * x), self.exitrates)
        ).item()

    def getcumprob(self, x: float) -> float:
        """
        Returns the cumulative distribution function P(X ≤ x).

        Args:
            x (float): The time at which the cumulative probability is evaluated.

        Returns:
            float: The cumulative probability.
        """
        return 1 - np.sum(np.matmul(self.pi, expm(self.phgen * x)))

    def getloglik(self) -> float:
        """
        Returns the log-likelihood of the fitted model.

        Args:
            None

        Returns:
            float: The log-likelihood value.
        """
        return self.loglikelihood

    def getaic(self) -> float:
        """
        Returns Akaike's Information Criterion (AIC).

        Args:
            None

        Returns:
            float: The AIC value.
        """
        return -2 * self.loglikelihood + 2 * self.nparam

    def getbic(self) -> float:
        """
        Returns the Bayesian Information Criterion (BIC).

        Args:
            None

        Returns:
            float: The BIC value.
        """
        return -2 * self.loglikelihood + self.nparam * np.log(len(self.obs))

    def __initialize(self) -> None:
        """
        Initializes internal parameters and prepares the model for fitting.

        Args:
            None

        Returns:
            None
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        self.obs = self.obs.astype(float)
        self.initpi = self.initpi.astype(float)
        self.initphgen = self.initphgen.astype(float)
        self.initexitrates = self.initexitrates.astype(float)
        self.identity = np.eye(self.nphases)

        self.pi = np.array(self.initpi, dtype=float).flatten()
        self.phgen = self.initphgen
        self.exitrates = np.array(self.initexitrates, dtype=float).flatten()

        if self.randominit:
            self.__initrandom()

        self.__countParameters()

    def __initrandom(self) -> None:
        """
        Randomly initializes the parameters of the phase-type distribution.

        Args:
            None

        Returns:
            None
        """

        nzidx = np.nonzero(self.pi)
        u = np.random.uniform(low=0.0, high=1.0, size=len(nzidx))
        u = u / np.sum(u)
        self.pi[nzidx] = u

        nzidx = np.nonzero(self.exitrates)
        u = np.random.uniform(low=0.0, high=1.0, size=len(nzidx))
        self.exitrates[nzidx] = u

        for i in range(self.nphases):
            nzidx = np.nonzero(np.ravel(self.phgen[i, :]))[0]
            msk = nzidx != i
            nzidx = nzidx[msk]
            u = np.random.uniform(low=0.0, high=1.0, size=len(nzidx))
            self.phgen[i, nzidx] = u
            self.phgen[i, i] = -(np.sum(u) + self.exitrates[i])

    def __estep(self) -> None:
        """
        Performs the expectation (E) step of the EM algorithm.

        Args:
            None

        Returns:
            None
        """
        self.bi = np.zeros(self.nphases)
        self.zi = np.zeros(self.nphases)
        self.ni = np.zeros(self.nphases)
        self.nij = np.zeros((self.nphases, self.nphases))
        self.loglikelihood = 0.0

        for y in self.obs:
            self.__Jmatrix(y)
            eTyt = np.matmul(self.eTy, self.exitrates)
            pieTy = np.matmul(self.pi, self.eTy)
            pieTyt = np.matmul(pieTy, self.exitrates)
            self.loglikelihood += np.log(pieTyt)
            for i in range(self.nphases):
                self.bi[i] += (self.pi[i] * eTyt[i]) / pieTyt
                self.zi[i] += self.Jmat[i, i] / pieTyt
                for j in range(self.nphases):
                    if j != i:
                        self.nij[i, j] += (self.phgen[i, j] * self.Jmat[j, i]) / pieTyt
                self.ni[i] += (pieTy[i] * self.exitrates[i]) / pieTyt

    def __mstep(self) -> None:
        """
        Performs the maximization (M) step of the EM algorithm.

        Args:
            None

        Returns:
            None
        """
        self.pi = self.bi / self.obs.size

        self.exitrates = self.ni / self.zi

        for i in range(self.nphases):
            for j in range(self.nphases):
                if j != i:
                    self.phgen[i, j] = self.nij[i, j] / self.zi[i]
            off_diag_sum = np.sum([self.phgen[i, j] for j in range(self.nphases) if j != i])
            self.phgen[i, i] = -(off_diag_sum + self.exitrates[i])


    def __updatelikelihood(self) -> None:
        """
        Updates the log-likelihood based on current model parameters.

        Args:
            None

        Returns:
            None
        """
        self.loglikelihood = 0.0
        for y in self.obs:
            self.loglikelihood += np.log(self.getdensity(y))

    def __Jmatrix(self, y: float) -> None:
        """
        Computes the J matrix and matrix exponential exp(Ty).

        Args:
            y (float): Observation value.

        Returns:
            None
        """

        t = self.exitrates[:, None]
        pi = self.pi[None, :]

        mat = expm(
            np.block([
                [self.phgen, np.matmul(t,pi)],
                [np.zeros((self.nphases, self.nphases)), self.phgen],
            ]) * y
        )

        self.eTy = mat[: self.nphases, : self.nphases]
        self.Jmat = mat[: self.nphases, self.nphases : 2 * self.nphases]


    def __countParameters(self) -> None:
        """
        Counts the number of independent model parameters.

        Args:
            None

        Returns:
            None
        """
        phg = 0
        for i in range(self.nphases):
            phg += (
                np.count_nonzero(self.phgen[i, :])
                + np.count_nonzero(self.exitrates[i])
                - 1
            )
        self.nparam = phg + (np.count_nonzero(self.pi) - 1)

    def __polish(self) -> None:
        """
        Normalizes model parameters.

        Args:
            None

        Returns:
            None
        """
        self.pi = self.pi / np.sum(self.pi)


import sys
import numpy as np
from scipy.linalg import expm
from scipy.stats import lognorm, norm, gamma, weibull_min, chi2
import matplotlib.pyplot as plt
from phasedist.dist import dist


class fitcph2dist:
    """
    Fit a continuous-time phase-type distribution to
    a distribution with a continuous density using
    the EM algorithm from p. 681 Bladt and Nielsen (2017).

    References:
        Bladt, M., & Nielsen, B. F. (2017). Matrix-Exponential Distributions in Applied Probability.
        Springer. https://doi.org/10.1007/978-1-4939-7049-0
    """

    def __init__(
        self,
        nphases: int = 2,
        dtype: str = "general",
        initdist: np.array = None,
        initphgen: np.array = None,
        initexitrates: np.array = None,
        randominit: bool = True,
        seed: int = None,
        tolerance: float = 1e-3,
        truncation: float = 0.99,
        steps: int = 50,
        itermax: int = 1000000000,
        verbose: bool = False,
    ) -> None:
        """
        Initializes the phase-type distribution fitting object.
        
        Args:
            nphases (int): Number of phases.
            dtype (str): Distribution structure type.
            initdist (np.array): Initial distribution vector.
            initphgen (np.array): Initial generator matrix.
            initexitrates (np.array): Initial exit rates.
            randominit (bool): Whether to randomly initialize parameters.
            seed (int): Random seed.
            tolerance (float): Convergence tolerance.
            truncation (float): Truncation level.
            steps (int): Number of integration steps.
            itermax (int): Maximum number of iterations.
            verbose (bool): Verbosity flag.
            
        Returns:
            None
        """
        
        self.nphases = nphases
        self.dtype = dtype
        self.initdist = initdist
        self.initphgen = initphgen
        self.initexitrates = initexitrates
        self.randominit = randominit
        self.seed = seed
        self.itermax = itermax
        self.tolerance = tolerance
        self.truncation = truncation
        self.disttype = None
        self.verbose = verbose
        self.dist = None
        self.steps = steps  # number of steps in the numerical integration

        # checking and fitting
        if self.__checkinputs():
            self.__makedist()  # fit the parameters
        else:
            sys.exit(1)  # terminate the program

    # ----------------------------------------------------------------------
    #   PUBLIC METHODS
    # ----------------------------------------------------------------------

    def lognorm(self, mu: float = None, sigma: float = None, mean: float = None, var: float = None) -> None:
        """
        Configures approximation to a log-normal distribution.
        
        Args:
            mu (float): Mean of underlying normal distribution.
            sigma (float): Standard deviation of underlying normal distribution.
            mean (float): Mean of the log-normal distribution.
            var (float): Variance of the log-normal distribution.
        
        Returns:
            None
        """     
        if mu is None and mean is not None:
            if mean <= 0:
                print("Error: 'mean<=0' is infeasible for the lognormal distribution.")
                sys.exit(1)
            self.param1 = np.log(np.power(mean, 2) / np.sqrt(np.power(mean, 2) + var))
            self.param2 = np.log(1 + var / np.power(mean, 2))
        elif mu is not None and mean is None:
            self.param1 = mu
            self.param2 = np.power(sigma, 2)
        self.disttype = "lognorm"
        self.__initialize()

    def norm(self, mu: float = None, sigma: float = None) -> None:
        """
        Configures approximation to a truncated normal distribution.
        
        Args:
            mu (float): Mean of the normal distribution.
            sigma (float): Standard deviation of the normal distribution.
        
        Returns:
            None
        """
        self.param1 = mu
        self.param2 = sigma
        self.disttype = "norm"
        self.__initialize()

    def gamma(self, shape: float = None, scale: float = None, rate: float = None) -> None:
        """
        Configures approximation to a gamma distribution.
        
        Args:
            shape (float): Shape parameter.
            scale (float): Scale parameter.
            rate (float): Rate parameter.
        
        Returns:
            None
        """
        self.param1 = shape
        if scale is None:
            self.param2 = 1 / rate
        elif rate is None:
            self.param2 = scale
        self.disttype = "gamma"
        self.__initialize()

    def chisq(self, df: int = None) -> None:
        """
        Configures approximation to a chi-squared distribution.
        
        Args:
            df (int): Degrees of freedom.
        
        Returns:
            None
        """
        self.param1 = df
        self.disttype = "chisq"
        self.__initialize()

    def weibull(self, shape: float = None, scale: float = None) -> None:
        """
        Configures approximation to a Weibull distribution.
        
        Args:
            shape (float): Shape parameter.
            scale (float): Scale parameter.
        
        Returns:
            None
        """
        self.param1 = shape
        self.param2 = scale
        self.disttype = "weibull"
        self.__initialize()

    def phasedist(self, initdist: np.array, phgen: np.array) -> None:
        """
        Configures approximation to an existing phase-type distribution.
        
        Args:
            initdist (np.array): Initial distribution vector.
            phgen (np.array): Generator matrix.
        
        Returns:
            None
        """
        self.param1 = initdist
        self.param2 = phgen
        self.disttype = "ph"
        self.__initialize()

    def percentiles(self, cumprobs: np.array = None, x: np.array = None) -> None:
        """
        Configures approximation using (empirical) cumulative probabilities.
        
        Args:
            cumprobs (np.array): Cumulative probabilities.
            x (np.array): Corresponding values.
        
        Returns:
            None
        """
        self.param1 = cumprobs
        self.param2 = x
        self.disttype = "per"
        self.__initialize()

    def fit(self) -> int:
        """
        Fits the phase-type distribution using the EM algorithm.
        
        Args:
            None
        
        Returns:
            int: Status code.
        """
        if self.disttype is None:
            print("Error: Select a distribution for the approximation.")
            return 1
        iter = 0
        self.eps = np.inf
        self.pi0 = np.copy(self.pi)
        self.phgen0 = np.copy(self.phgen)
        while iter < self.itermax and self.eps > self.tolerance:
            self.__estep()
            self.__mstep()
            self.__updateEpsilon()
            self.pi0 = np.copy(self.pi)
            self.phgen0 = np.copy(self.phgen)
            iter += 1
            if self.verbose and iter % 5 == 0:
                d = dist(discrete=False, initdist=self.pi, phgen=self.phgen)
                print(
                    "iter =",
                    iter,
                    "  eps =",
                    self.eps,
                    "  mean =",
                    d.getmean(),
                    "  var =",
                    d.getvar(),
                )

        # create object for output PH distribution
        self.dist = dist(
            discrete=False, initdist=self.pi, phgen=self.phgen, seed=self.seed
        )

        return 0

    def plot(self) -> None:
        """
        Plots the fitted distribution and target distribution.
        
        Args:
            None
        
        Returns:
            None
        """
        x = np.linspace(0.0, self.y.max(), 500)
        dist_pdf = np.zeros(len(x))
        ph_pdf = np.zeros(len(x))
        for i in range(len(x)):
            if self.disttype == "lognorm":
                dist_pdf[i] = lognorm.pdf(x[i], self.param2, scale=np.exp(self.param1))
            elif self.disttype == "gamma":
                dist_pdf[i] = gamma.pdf(x[i], self.param1, scale=self.param2)
            elif self.disttype == "weibull":
                dist_pdf[i] = weibull_min.pdf(x[i], self.param1, scale=self.param2)
            elif self.disttype == "chisq":
                dist_pdf[i] = chi2.pdf(x[i], self.param1)
            elif self.disttype == "ph":
                d = dist(discrete=False, initdist=self.param1, phgen=self.param2)
                dist_pdf[i] = d.getdensity(x[i])
            ph_pdf[i] = self.getdensity(x[i])

        # make plot
        plt.figure(figsize=(10, 6))
        if not self.disttype == "norm" and not self.disttype == "per":
            plt.plot(x, dist_pdf, label="True density", color="blue")
        plt.plot(x, ph_pdf, label="Approx. density", color="red", linestyle="--")
        plt.xlabel("x")
        plt.ylabel("Density")
        plt.title("Approximation validation")
        plt.legend()
        plt.grid(True)
        plt.show()

        return None

    def getinitdist(self) -> np.array:
        """
        Returns the initial distribution vector.
        
        Args:
            None
        
        Returns:
            np.array: Initial distribution.
        """
        if self.dist is not None:
            return self.dist.getinitdist()
        else:
            return np.nan

    def getphasegen(self) -> np.array:
        """
        Returns the phase generator matrix.
        
        Args:
            None
        
        Returns:
            np.array: Generator matrix.
        """
        if self.dist is not None:
            return self.dist.getphasegen()
        else:
            return np.nan

    def getexitrates(self) -> np.array:
        """
        Returns the exit rate vector.
        
        Args:
            None
        
        Returns:
            np.array: Exit rates.
        """
        if self.dist is not None:
            return self.dist.getexitrates()
        else:
            return np.nan

    def getmean(self) -> float:
        """
        Returns the mean of the fitted distribution.
        
        Args:
            None
        
        Returns:
            float: Mean.
        """
        if self.dist is not None:
            return self.dist.getmean()
        else:
            return np.nan

    def getvar(self) -> float:
        """
        Returns the variance of the fitted distribution.
        
        Args:
            None
        
        Returns:
            float: Variance.
        """
        if self.dist is not None:
            return self.dist.getvar()
        else:
            return np.nan

    def getdensity(self, x: float) -> float:
        """
        Returns the distribution's density, f(x).
        
        Args:
            x (float): The distribution's density will be computed at time of x.
        
        Returns:
            float: The computed density.
        """
        if self.dist is not None:
            return self.dist.getdensity(x)
        else:
            return np.nan

    def getcumprob(self, x: float) -> float:
        """
        Returns the cumulative distribution value at x.
        
        Args:
            x (float): Evaluation point.
        
        Returns:
            float: Cumulative probability.
        """
        if self.dist is not None:
            return self.dist.getcumprob(x)
        else:
            return np.nan

    def getquantile(self, p: float, tolerance: float = 1e-9) -> float:
        """
        Returns the quantile corresponding to probability p.
        
        Args:
            p (float): Probability level.
            tolerance (float): Numerical tolerance.
        
        Returns:
            float: Quantile value.
        """
        if self.dist is not None:
            return self.dist.getquantile(p, tolerance)
        else:
            return np.nan

    def getdist(self) -> dist:
        """
        Returns the fitted phase-type distribution object.
        
        Args:
            None
        
        Returns:
            dist: Phase-type distribution.
        """
        return self.dist

    # ----------------------------------------------------------------------
    #   PRIVATE METHODS
    # ----------------------------------------------------------------------

    def __checkinputs(self) -> bool | int:
        """
        Checks feasibility and validity of all input parameters.
        
        Args:
            None
        
        Returns:
            bool | int: True if inputs are valid, False or 0 otherwise.
        """
        if not isinstance(self.nphases, int) or self.nphases < 1:
            print(
                "Error: The number of phases can only be specified as an integer larger than 0."
            )
            return 0
        if not isinstance(self.dtype, str):
            print("Error: The distribution type can only be specified as a string.")
        if self.initdist is not None and (
            isinstance(self.initdist, np.ndarray) or isinstance(self.initdist, list)
        ):
            self.initdist = np.matrix(self.initdist)
        elif self.initdist is not None and not isinstance(self.initdist, np.matrix):
            print(
                "Error: The initial distribution can only be specified as a list, NumPy array, or a NumPy matrix."
            )
            return False
        if self.initphgen is not None and (
            isinstance(self.initphgen, np.ndarray) or isinstance(self.initphgen, list)
        ):
            self.initphgen = np.matrix(self.initphgen)
        elif self.initphgen is not None and not isinstance(self.initphgen, np.matrix):
            print(
                "Error: The PH generator can only be specified as a list or a NumPy matrix."
            )
            return False
        if self.initexitrates is not None and (
            isinstance(self.initexitrates, np.ndarray)
            or isinstance(self.initexitrates, list)
        ):
            self.initexitrates = np.transpose(np.matrix(self.initexitrates))
        elif self.initexitrates is not None and not isinstance(
            self.initexitrates, np.matrix
        ):
            print(
                "Error: The exit rate vector can only be specified as a list, NumPy array, or a NumPy matrix."
            )
            return False
        if not isinstance(self.randominit, bool):
            print("Error: The argument 'randominit' needs to be of type 'bool'.")
            return False
        if self.seed is not None and not isinstance(self.seed, int):
            print("Error: The seed can only be specified as an integer.")
            return False
        if not isinstance(self.tolerance, float):
            print("Error: The argument 'tolerance' needs to be of type 'float'.")
            return False
        if not isinstance(self.itermax, float) and not isinstance(self.itermax, int):
            print("Error: The argument 'itermax' needs to be of type 'float' or 'int'.")
            return False
        if not isinstance(self.verbose, bool):
            print("Error: The argument 'verbose' needs to be of type 'bool'.")
            return False

        # check PH generator and exit rates in case of no random initialization
        if self.dtype == "custom" or not self.randominit:
            self.nphases = self.initphgen.shape[0]
        if not self.randominit:
            if not self.__correctphgen(
                self.initphgen, self.initexitrates
            ) or not self.__correctinitdist(self.d.getinitdist):
                return False

        return True

    def __correctphgen(self, phasegen: np.array, exitrates: np.array) -> bool:
        """
        Checks feasibility of the PH generator and exit rate vector.
        
        Args:
            phasegen (np.array): Phase-type generator matrix.
            exitrates (np.array): Exit rate vector.
        
        Returns:
            bool: True if feasible, False otherwise.
        """

        # check PH generator
        if phasegen.shape[0] != phasegen.shape[1] or phasegen.shape[0] != self.nphases:
            print(
                "Error: The dimensions of the PH generator does not match the number of phases."
            )
            return False
        if (
            np.where(np.isnan(phasegen))[0].size > 0
            or np.where(np.isinf(phasegen))[0].size > 0
            or np.where(np.isneginf(phasegen))[0].size > 0
        ):
            print("Error: The PH generator contains NaN or/and infinity values.")
            return False
        if np.any(phasegen[~np.eye(self.nphases, dtype=bool)] < 0):
            print("Error: The PH generator contains negative off-diagonal values.")
            return False
        if np.any(phasegen[np.eye(self.nphases, dtype=bool)] > 0):
            print("Error: The PH generator contains positive diagonal values.")
            return False
        if np.max(abs(np.add(np.sum(phasegen, axis=1), exitrates))) > 1e-6:
            print(
                "Warning: An element of the exit rate vector deviates at least 1e-6 from the absolute row sum of the PH generator."
            )
            return True
        # check exit rates
        if self.nphases != exitrates.size:
            print(
                "Error: The size of the exit rate vector does not match the number of phases."
            )
            return False
        if (
            np.where(np.isnan(exitrates))[0].size > 0
            or np.where(np.isinf(exitrates))[0].size > 0
            or np.where(np.isneginf(exitrates))[0].size > 0
        ):
            print("Error: The exit rate vector contains NaN or/and infinity values.")
            return False
        if np.any(exitrates < 0):
            print("Error: The exit rate vector contains negative values.")
            return False
        return True

    def __correctinitdist(self, initdist: np.array) -> bool:
        """
        Checks feasibility of the initial distribution.
        
        Args:
            initdist (np.array): Initial distribution vector.
        
        Returns:
            bool: True if feasible, False otherwise.
        """
        if initdist.size != self.nphases:
            print(
                "Error: The size of the initial distribution does not match the number of phases."
            )
            return False
        if (
            np.where(np.isnan(initdist))[0].size > 0
            or np.where(np.isinf(initdist))[0].size > 0
            or np.where(np.isneginf(initdist))[0].size > 0
        ):
            print(
                "Error: The initial distribution contains NaN or/and infinity values."
            )
            return False
        if np.any(initdist < 0.0):
            print("Error: The initial distribution contains negative values.")
            return False
        if np.abs(np.sum(initdist) - 1.0) > 1e-14:
            print(
                "Warning: Prior to adjusting for zeros in the observations the initial distribution summed to "
                + str(np.sum(initdist))
            )
            return False
        return True

    def __makedist(self) -> None | int:
        """
        Initializes distribution structure based on the specified type.
        
        Args:
            None
        
        Returns:
            None | int: None if successful, error code otherwise.
        """
        if self.dtype == "general":
            self.__general()
        elif self.dtype == "generlang":
            self.__generlang()
        elif self.dtype == "hyperexp":
            self.__hyperexp()
        elif self.dtype == "coxian":
            self.__coxian()
        elif self.dtype == "gencoxian":
            self.__gencoxian()
        elif self.dtype != "custom":
            print("Error: Unknown distribution type.")
            return 1

    def __general(self) -> None:
        """
        Initializes parameters for a general phase-type distribution.
        
        Args:
            None
        
        Returns:
            None
        """
        self.initdist = np.matrix(np.ones((1, self.nphases)))
        self.initphgen = np.matrix(np.ones((self.nphases, self.nphases)))
        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))

    def __generlang(self) -> None:
        """
        Initializes parameters for a generalized Erlang distribution.
        
        Args:
            None
        
        Returns:
            None
        """
        self.initdist = np.matrix(np.zeros((1, self.nphases)))
        self.initdist[0, 0] = 1

        self.initexitrates = np.matrix(np.zeros((self.nphases, 1)))
        self.initexitrates[self.nphases - 1, 0] = 1

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1
            if i < (self.nphases - 1):
                self.initphgen[i, i + 1] = 1

    def __hyperexp(self) -> None:
        """
        Initializes parameters for a hyper-exponential distribution.
        
        Args:
            None
        
        Returns:
            None
        """
        self.initdist = np.matrix(np.ones((1, self.nphases)))

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1

    def __coxian(self) -> None:
        """
        Initializes parameters for a Coxian distribution.
        
        Args:
            None
        
        Returns:
            None
        """
        self.initdist = np.matrix(np.zeros((1, self.nphases)))
        self.initdist[0, 0] = 1

        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1
            if i < (self.nphases - 1):
                self.initphgen[i, i + 1] = 1

    def __gencoxian(self) -> None:
        """
        Initializes parameters for a generalized Coxian distribution.
        
        Args:
            None
        
        Returns:
            None
        """
        self.initdist = np.matrix(np.ones((1, self.nphases)))
        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1
            if i < (self.nphases - 1):
                self.initphgen[i, i + 1] = 1

    def __initialize(self) -> None:
        """
        Initializes parameters, random values, and numerical integration grid.
        
        Args:
            None
        
        Returns:
            None
        """        
        if self.seed is not None:
            np.random.seed(self.seed)

        # convert data types
        self.steps = int(self.steps)
        self.initpi = self.initdist.astype(float)
        self.initphgen = self.initphgen.astype(float)
        self.initexitrates = self.initexitrates.astype(float)

        # copy to output parameters
        self.pi = self.initdist
        self.phgen = self.initphgen
        self.exitrates = self.initexitrates

        # initialize with random parameters
        # accounting for the specified structure
        if self.randominit:
            self.__initrandom()

        # create the cumulated probabilities and evaluation points
        self.__computeyvector()

    def __initrandom(self) -> None:
        """
        Randomly initializes parameters of a continuous-time PH distribution.
        
        Args:
            None
        
        Returns:
            None
        """                
        
        # make a random initial distribution (pi)
        nzidx = np.nonzero(self.pi)[1]
        u = np.random.uniform(low=0.0, high=1.0, size=len(nzidx))
        u = u / np.sum(u)
        self.pi[0, nzidx] = u

        # make a random exit vector
        nzidx = np.nonzero(self.exitrates)[0]
        u = np.random.uniform(low=0.0, high=10.0, size=len(nzidx))
        self.exitrates[nzidx, 0] = u

        # make a random PH generator
        for i in range(self.nphases):
            nzidx = np.nonzero(self.phgen[i, :])[1]
            msk = nzidx != i
            nzidx = nzidx[msk]
            u = np.random.uniform(low=0.0, high=10.0, size=len(nzidx))
            self.phgen[i, nzidx] = u
            self.phgen[i, i] = -(np.sum(u) + self.exitrates[i, 0])

    def __estep(self) -> None:
        """
        Performs the expectation step of the EM algorithm.
        
        Args:
            None
        
        Returns:
            None
        """
        self.bi = np.zeros(self.nphases)
        self.zi = np.zeros(self.nphases)
        self.ni = np.zeros(self.nphases)
        self.nij = np.zeros((self.nphases, self.nphases))

        # pre-compute eTy and pieTyt
        self.eTy = [None] * self.steps
        self.eTyut = [[None] * (self.steps + 1) for _ in range(self.steps)]
        self.pieTu = [[None] * (self.steps + 1) for _ in range(self.steps)]
        self.pieTyt = np.zeros(self.steps)
        for k in range(self.steps):
            self.eTy[k] = expm(self.phgen * self.y[k])
            self.pieTyt[k] = np.matmul(self.pi, np.matmul(self.eTy[k], self.exitrates))

        for i in range(self.nphases):

            # bi
            for k in range(self.steps):
                eTyt = np.matmul(self.eTy[k], self.exitrates)
                Gy = (self.pi[0, i] * eTyt[i, 0]) / self.pieTyt[k]
                self.bi[i] += Gy * self.hy[k]

            # zi (denominator used in calculation of PH generator and exit rates)
            for k in range(self.steps):
                # inner integral
                innerint = 0.0
                if self.y[k] > 0:
                    u = np.linspace(0, self.y[k], self.steps + 1)
                    for l in range(0, len(u) - 2, 2):
                        # inner fa
                        self.pieTu[k][l] = np.matmul(self.pi, expm(self.phgen * u[l]))
                        self.eTyut[k][l] = np.matmul(
                            expm(self.phgen * (self.y[k] - u[l])), self.exitrates
                        )
                        inner_fa = self.pieTu[k][l][0, i] * self.eTyut[k][l][i, 0]
                        # inner fmid
                        self.pieTu[k][l + 1] = np.matmul(
                            self.pi, expm(self.phgen * u[l + 1])
                        )
                        self.eTyut[k][l + 1] = np.matmul(
                            expm(self.phgen * (self.y[k] - u[l + 1])), self.exitrates
                        )
                        inner_fmid = (
                            self.pieTu[k][l + 1][0, i] * self.eTyut[k][l + 1][i, 0]
                        )
                        # inner fb
                        self.pieTu[k][l + 2] = np.matmul(
                            self.pi, expm(self.phgen * u[l + 2])
                        )
                        self.eTyut[k][l + 2] = np.matmul(
                            expm(self.phgen * (self.y[k] - u[l + 2])), self.exitrates
                        )
                        inner_fb = (
                            self.pieTu[k][l + 2][0, i] * self.eTyut[k][l + 2][i, 0]
                        )

                        innerint += self.__simpsonsrule(
                            u[l], u[l + 2], inner_fa, inner_fmid, inner_fb
                        )
                # outer integral
                Gy = innerint / self.pieTyt[k]
                self.zi[i] += Gy * self.hy[k]

            # ni (numerator used in calculation of exit rates)
            for k in range(self.steps):
                pieTy = np.matmul(self.pi, self.eTy[k])
                Gy = (pieTy[0, i] / self.pieTyt[k]) * self.exitrates[i, 0]
                self.ni[i] += Gy * self.hy[k]

            for j in range(self.nphases):
                # nij (numerator used in calculation of PH generator)
                if j != i and self.phgen[i, j] > 0.0:
                    for k in range(self.steps):
                        # inner integral
                        innerint = 0.0
                        if self.y[k] > 0:
                            u = np.linspace(0, self.y[k], self.steps + 1)
                            for l in range(0, len(u) - 2, 2):
                                # inner fa
                                inner_fa = (
                                    self.pieTu[k][l][0, i] * self.eTyut[k][l][j, 0]
                                )
                                # inner fmid
                                inner_fmid = (
                                    self.pieTu[k][l + 1][0, i]
                                    * self.eTyut[k][l + 1][j, 0]
                                )
                                # inner fb
                                inner_fb = (
                                    self.pieTu[k][l + 2][0, i]
                                    * self.eTyut[k][l + 2][j, 0]
                                )

                                innerint += self.__simpsonsrule(
                                    u[l], u[l + 2], inner_fa, inner_fmid, inner_fb
                                )
                        # outer integral
                        Gy = (self.phgen[i, j] / self.pieTyt[k]) * innerint
                        self.nij[i, j] += Gy * self.hy[k]

    def __mstep(self) -> None:
        """
        Performs the maximization step of the EM algorithm.
        
        Args:
            None
        
        Returns:
            None
        """

        # update the initial distribution
        for i in range(self.nphases):
            self.pi[0, i] = self.bi[i]
        self.pi = self.pi / np.sum(self.pi)

        # update the exit rates
        for i in range(self.nphases):
            self.exitrates[i, 0] = self.ni[i] / self.zi[i]

        # update the PH generator
        for i in range(self.nphases):
            sm = self.exitrates[i, 0]
            for j in range(self.nphases):
                if j != i:
                    self.phgen[i, j] = self.nij[i, j] / self.zi[i]
                    sm += self.phgen[i, j]
            self.phgen[i, i] = -sm

    def __updateEpsilon(self):
        """
        Computes the maximum absolute change in parameters.
        
        Args:
            None
        
        Returns:
            None
        """        
        # eps1 = np.max(np.abs(np.divide(np.subtract(self.pi[np.nonzero(self.pi0)],self.pi0[np.nonzero(self.pi0)]),self.pi0[np.nonzero(self.pi0)])))
        # eps2 = np.max(np.abs(np.divide(np.subtract(self.phgen[np.nonzero(self.phgen0)],self.phgen0[np.nonzero(self.phgen0)]),self.phgen0[np.nonzero(self.phgen0)])))
        eps1 = np.max(
            np.abs(
                np.subtract(
                    self.pi[np.nonzero(self.pi0)], self.pi0[np.nonzero(self.pi0)]
                )
            )
        )
        eps2 = np.max(
            np.abs(
                np.subtract(
                    self.phgen[np.nonzero(self.phgen0)],
                    self.phgen0[np.nonzero(self.phgen0)],
                )
            )
        )
        self.eps = np.max(np.array([eps1, eps2]))

    def __computeyvector(self) -> None:
        """
        Computes evaluation points and probability weights for numerical integration.
        
        Args:
            None
        
        Returns:
            None
        """

        # make evaluation points
        if self.disttype == "lognorm":
            self.y = np.linspace(
                0,
                lognorm.ppf(self.truncation, self.param2, scale=np.exp(self.param1)),
                self.steps + 1,
            )
        elif self.disttype == "gamma":
            self.y = np.linspace(
                0,
                gamma.ppf(self.truncation, self.param1, scale=self.param2),
                self.steps + 1,
            )
        elif self.disttype == "norm":
            self.y = np.linspace(
                0, self.__normtruncquantfun(self.param1, self.param2), self.steps + 1
            )
        elif self.disttype == "weibull":
            self.y = np.linspace(
                0,
                weibull_min.ppf(self.truncation, self.param1, scale=self.param2),
                self.steps + 1,
            )
        elif self.disttype == "chisq":
            self.y = np.linspace(
                0, chi2.ppf(self.truncation, self.param1), self.steps + 1
            )
        elif self.disttype == "ph":
            d = dist(discrete=False, initdist=self.param1, phgen=self.param2)
            self.y = np.linspace(0, d.getquantile(self.truncation), self.steps + 1)
        elif self.disttype == "per":
            self.y = np.linspace(0, np.max(self.param2), self.steps + 1)

        # compute cumulated probability segments
        self.hy = np.zeros(self.steps)
        for i in range(self.steps):
            if self.disttype == "lognorm":
                self.hy[i] = self.__lognorm_dcdf(self.y[i], self.y[i + 1])
            elif self.disttype == "gamma":
                self.hy[i] = self.__gamma_dcdf(self.y[i], self.y[i + 1])
            elif self.disttype == "norm":
                self.hy[i] = self.__normtrunc_dcdf(self.y[i], self.y[i + 1])
            elif self.disttype == "weibull":
                self.hy[i] = self.__weib_dcdf(self.y[i], self.y[i + 1])
            elif self.disttype == "chisq":
                self.hy[i] = self.__chisq_dcdf(self.y[i], self.y[i + 1])
            elif self.disttype == "ph":
                self.hy[i] = self.__ph_dcdf(self.y[i], self.y[i + 1])
            elif self.disttype == "per":
                self.hy[i] = self.__per_dcdf(self.y[i], self.y[i + 1])

        # adjust y's for the E-step
        self.y = self.y[1:]

    def __simpsonsrule(self, a: float, b: float, fa: float, fmid: float, fb: float) -> float:
        """
        Computes Simpson’s 1/3 rule approximation.
        
        Args:
            a (float): Lower bound.
            b (float): Upper bound.
            fa (float): Function value at a.
            fmid (float): Function value at midpoint.
            fb (float): Function value at b.
        
        Returns:
            float: Approximated integral value.
        """
        # note: fmid = f((a+b)/2)
        return ((b - a) / 6) * (fa + 4 * fmid + fb)

    def lognormdensity(self, x: float) -> float:
        """
        Probability density function of the log-normal distribution.
        
        Args:
            x (float): Compute the density at x.
        
        Returns:
            float: Density.
        """        
        return (1 / (x * np.sqrt(self.param2) * np.sqrt(2 * np.pi))) * np.exp(
            -(np.power(np.log(x) - self.param1, 2) / (2 * self.param2))
        )

    def __lognorm_dcdf(self, x0: float, x1: float) -> float:
        """
        Computes cumulative probability between two points for a log-normal distribution.
        
        Args:
            x0 (float): Lower bound.
            x1 (float): Upper bound.
        
        Returns:
            float: Cumulative probability.
        """
        if x0 == 0:
            return norm.cdf((np.log(x1) - self.param1) / np.sqrt(self.param2))
        else:
            return norm.cdf(
                (np.log(x1) - self.param1) / np.sqrt(self.param2)
            ) - norm.cdf((np.log(x0) - self.param1) / np.sqrt(self.param2))

    def __gamma_dcdf(self, x0: float, x1: float) -> float:
        """
        Computes cumulative probability between two points for a gamma distribution.
        
        Args:
            x0 (float): Lower bound.
            x1 (float): Upper bound.
        
        Returns:
            float: Cumulative probability.
        """
        if x0 == 0:
            return gamma.cdf(x1, self.param1, scale=self.param2)
        else:
            return gamma.cdf(x1, self.param1, scale=self.param2) - gamma.cdf(
                x0, self.param1, scale=self.param2
            )

    def __normtrunc_dcdf(self, x0: float, x1: float) -> float:
        """
        Computes cumulative probability between two points for a truncated normal distribution.
        
        Args:
            x0 (float): Lower bound.
            x1 (float): Upper bound.
        
        Returns:
            float: Cumulative probability.
        """
        return (
            norm.cdf((x1 - self.param1) / self.param2)
            - norm.cdf((x0 - self.param1) / self.param2)
        ) / (1 - norm.cdf((-self.param1) / self.param2))

    def __weib_dcdf(self, x0: float, x1: float) -> float:
        """
        Computes cumulative probability between two points for a Weibull distribution.
        
        Args:
            x0 (float): Lower bound.
            x1 (float): Upper bound.
        
        Returns:
            float: Cumulative probability.
        """
        if x0 == 0:
            return weibull_min.cdf(x1, self.param1, scale=self.param2)
        else:
            return weibull_min.cdf(
                x1, self.param1, scale=self.param2
            ) - weibull_min.cdf(x0, self.param1, scale=self.param2)

    def __chisq_dcdf(self, x0: float, x1: float) -> float:
        """
        Computes cumulative probability between two points for a chi-square distribution.
        
        Args:
            x0 (float): Lower bound.
            x1 (float): Upper bound.
        
        Returns:
            float: Cumulative probability.
        """
        if x0 == 0:
            return chi2.cdf(x1, self.param1)
        else:
            return chi2.cdf(x1, self.param1) - chi2.cdf(x0, self.param1)

    def __ph_dcdf(self, x0: float, x1: float) -> float:
        """
        Computes cumulative probability between two points for a PH distribution.
        
        Args:
            x0 (float): Lower bound.
            x1 (float): Upper bound.
        
        Returns:
            float: Cumulative probability.
        """
        if x0 == 0:
            return 1 - np.sum(np.matmul(self.param1, expm(self.param2 * x1)))
        else:
            return np.sum(np.matmul(self.param1, expm(self.param2 * x0))) - np.sum(
                np.matmul(self.param1, expm(self.param2 * x1))
            )

    def __per_dcdf(self, x0: float, x1: float) -> float:
        """
        Computes cumulative probability between two points using empirical percentiles.
        
        Args:
            x0 (float): Lower bound.
            x1 (float): Upper bound.
        
        Returns:
            float: Cumulative probability.
        """
        if x0 == 0:
            return self.__per_cdf(x1)
        else:
            return self.__per_cdf(x1) - self.__per_cdf(x0)

    def __per_cdf(self, x: float) -> float:
        """
        Computes cumulative probability at a point using empirical percentiles.
        
        Args:
            x (float): Evaluation point.
        
        Returns:
            float: Cumulative probability.
        """
        idx1 = np.min(np.where(self.param2 >= x))
        if idx1 > 0:
            idx0 = idx1 - 1
            return self.param1[idx0] + (self.param1[idx1] - self.param1[idx0]) * (
                (x - self.param2[idx0]) / (self.param2[idx1] - self.param2[idx0])
            )
        else:
            return self.param1[idx1] * (x / self.param2[idx1])

    def __normtruncquantfun(self, mu: float, sigma: float) -> float:
        """
        Computes numerical quantile for a truncated normal distribution.
        
        Args:
            mu (float): Mean of the normal distribution.
            sigma (float): Standard deviation.
        
        Returns:
            float: Quantile value.
        """
        cmp = 1 - norm.cdf(-mu / sigma)
        x = np.max(np.array([sigma * self.tolerance, mu]))
        trc = (norm.cdf((x - mu) / sigma) - norm.cdf(-mu / sigma)) / cmp
        dd = sigma * self.tolerance
        iter = 0
        while np.abs(trc - self.truncation) > self.tolerance and iter < self.itermax:
            x1 = x + dd
            f1 = (norm.cdf((x1 - mu) / sigma) - norm.cdf(-mu / sigma)) / cmp
            grad = (f1 - trc) / dd
            x = x - (trc - self.truncation) / grad
            trc = (norm.cdf((x - mu) / sigma) - norm.cdf(-mu / sigma)) / cmp
            iter += 1
        return x



import sys
import numpy as np
from scipy.linalg import expm
from scipy.stats import lognorm
import matplotlib.pyplot as plt
from typing import Union


class dist:
    """
    Phase-type distribution object that sets up a discrete or continuous phase-type (PH)
    distribution based on a provided initial distribution vector and
    phase-type generator matrix. The object supports evaluation of
    density, cumulative distribution, quantiles, and random sampling.
    """

    def __init__(
        self,
        discrete: bool = False,
        initdist: np.array = None,
        phgen: np.array = None,
        seed: int = None,
    ) -> None:
        """
        Initialize a phase-type distribution object.

        Args:
            discrete (bool, default=False):
                If True, the distribution is treated as a discrete
                phase-type (DPH) distribution. If False, it is treated as a
                continuous phase-type (CPH) distribution.
            initdist (np.array):
                The initial distribution vector (row vector). Must have
                length equal to the number of phases. If None, no explicit
                initial distribution is enforced.
            phgen (np.array):
                The PH generator matrix. For CPH this must be a sub-intensity
                matrix; for DPH, a sub-transition matrix. Its dimension
                defines the number of phases.
            seed (int):
                Random seed used for reproducible sampling.

        Notes
        -----
        Input validation is performed automatically. If validation fails,
        program execution terminates.
        """

        self.discrete = discrete
        self.initdist = initdist
        self.phgen = phgen
        self.nphases = self.phgen.shape[0]
        self.seed = seed

        if self.__checkinputs():  # check inputs
            self.__initialize()
        else:
            sys.exit(1)  # terminate the program

    def getinitdist(self) -> np.array:
        """
        Returns the initial distribution.

        Returns:
            np.array: The initial distribution (vector).
        """

        return self.initdist

    def getphasegen(self) -> np.array:
        """
        Returns the phase-type generator.

        Returns:
            np.array: The phase-type generator (matrix).
        """

        return self.phgen

    def getexitrates(self) -> np.array:
        """
        Returns the exitrate vector.

        Returns:
            np.array: The exit rate vector (vector).
        """

        return self.exitrates

    def getmean(self) -> float:
        """
        Returns the mean.

        Returns:
            float: The computed mean.
        """

        if self.discrete:
            return np.sum(
                np.matmul(
                    self.initdist,
                    np.linalg.inv(np.subtract(np.eye(self.nphases), self.phgen)),
                )
            )
        else:
            return -np.sum(np.matmul(self.initdist, np.linalg.inv(self.phgen)))

    def getvar(self) -> float:
        """
        Returns the variance.

        Returns:
            float: The computed variance.
        """

        if self.discrete:
            Tinv = np.linalg.inv(np.subtract(np.eye(self.nphases), self.phgen))
            return (
                np.sum(
                    np.matmul(
                        np.matmul(self.initdist, Tinv),
                        np.subtract((2 * Tinv), np.eye(self.nphases)),
                    )
                )
                - self.getmean() ** 2
            )
        else:
            phinv = np.linalg.inv(self.phgen)
            return 2 * np.sum(
                np.matmul(self.initdist, np.linalg.matrix_power(phinv, 2))
            ) - np.power(np.sum(np.matmul(self.initdist, phinv)), 2)

    def getdensity(self, x: float) -> float:
        """
        Returns the distribution's density, f(x).

        Args:
            x (float): The distribution's density will be computed at time of x.

        Returns:
            float: The computed density.
        """

        if isinstance(x, np.ndarray):
            y = np.zeros(x.size)
            for i in range(x.size):
                y[i] = self.__computedensity(x[i])
            return y
        else:
            return self.__computedensity(x)

    def getcumprob(self, x: float) -> float:
        """
        Returns the cumulated probability, P(X<=x).

        Args:
            x (float): The cumulated probability will be computed for the interval between 0 and x, i.e. P(X<=x).

        Returns:
            float: The computed probability.
        """

        if isinstance(x, np.ndarray):
            y = np.zeros(x.size)
            for i in range(x.size):
                y[i] = self.__computecumprob(x[i])
            return y
        else:
            return self.__computecumprob(x)

    def getquantile(self, p: float, tolerance: float = 1e-9) -> int | float:
        """
        Returns the x that ensures P(X<=x)=p, i.e. the quantile function of the
        PH distribution.

        Args:
            p (float): The cumulated probability.
            tolerance (float, default=1e-9): Tolerance for the numerical algorithm.

        Returns:
            int or float: The numerically evaluated quantile.
        """

        if isinstance(p, np.ndarray):
            y = np.zeros(p.size)
            for i in range(p.size):
                y[i] = self.__computequantile(p[i], tolerance)
            return y
        else:
            return self.__computequantile(p, tolerance)

    def getrandom(self, size: int = 1) -> int | float:
        """
        Samples a pseudo-random number from the distribution.

        Args:
            size (int, default=1): The sample size.

        Returns:
            int | float: The generated random number.
        """

        if size == 1:
            if self.discrete:
                return self.__dphsample()
            else:
                return self.__cphsample()
        elif size < 1:
            return np.nan
        else:
            obs = np.zeros(size)
            for i in range(size):
                if self.discrete:
                    obs[i] = self.__dphsample()
                else:
                    obs[i] = self.__cphsample()
            return obs

    def plot(self, type: str = "pdf") -> None:
        """
        Plots the density or cumulated distribution function.

        Args:
            type (str, default="pdf"): The plotted function ("pdf" or "cdf").

        Returns:
            None
        """

        # compute densities for approximate and true distributions
        x = np.linspace(0.0, self.getquantile(p=0.999), 500)
        val = np.zeros(len(x))
        for i in range(len(x)):
            if type == "pdf":
                val[i] = self.getdensity(x[i])
            elif type == "cdf":
                val[i] = self.getcumprob(x[i])

        if type == "pdf":
            ylbl = "Density"
            tl = "Probability Density Function"
        elif type == "cdf":
            ylbl = "Probability"
            tl = "Cumulative Distribution Function"

        # make plot
        plt.figure(figsize=(10, 6))
        plt.plot(x, val, label=ylbl, color="blue", linestyle="-")
        plt.xlabel("x")
        plt.ylabel(ylbl)
        plt.title(tl)
        plt.legend()
        plt.grid(True)
        plt.show()

        return None

    def __initialize(self) -> None:
        """
        Overall initialization.

        Returns:
            None
        """

        # compute exit rate vector
        if self.discrete:
            self.exitrates = 1 - np.sum(self.phgen, axis=1)
        else:
            self.exitrates = abs(np.sum(self.phgen, axis=1))

        # set a pre-defined seed
        if self.seed is not None:
            np.random.seed(self.seed)

        # prepare vectors for random sampling
        self.flatinitdist = np.asarray(self.initdist).ravel()
        self.a = []
        for s in range(self.nphases):
            if self.discrete:
                self.a.append(
                    np.append(
                        np.asarray(self.phgen[s, :]).ravel(), self.exitrates[s].item()
                    )
                )
            else:
                a = np.asarray(self.phgen[s, :] / (-self.phgen[s, s])).ravel()
                a[a < 0] = 0
                self.a.append(
                    np.append(a, (self.exitrates[s].item() / -self.phgen[s, s]))
                )

    def __checkinputs(self) -> bool:
        """
        Checks the feasibility of all input parameters.

        Returns:
            bool: Returns True if feasible; otherwise False.
        """

        # check data types and convert if necesarry
        if not isinstance(self.discrete, bool):
            print("Error: The argument 'discrete' needs to be of type 'bool'.")
        if self.initdist is not None and (
            isinstance(self.initdist, np.ndarray) or isinstance(self.initdist, list)
        ):
            self.initdist = np.matrix(self.initdist)
        elif self.initdist is not None and not isinstance(self.initdist, np.matrix):
            print(
                "Error: The initial distribution can only be specified as a list, NumPy array, or a NumPy matrix."
            )
            return False
        if self.phgen is not None and (
            isinstance(self.phgen, np.ndarray) or isinstance(self.phgen, list)
        ):
            self.phgen = np.matrix(self.phgen)
        elif self.phgen is not None and not isinstance(self.phgen, np.matrix):
            print(
                "Error: The PH generator can only be specified as a list or a NumPy matrix."
            )
            return False
        if self.seed is not None and not isinstance(self.seed, int):
            print("Error: The seed can only be specified as an integer.")
            return False
        return True  # if all correct

    def __computedensity(self, x: float) -> float:
        """
        Returns the distribution's density function.

        Args:
            x (float): Computes the density function for the scalar x.

        Returns:
            float: The computed density.
        """

        if self.discrete:
            if int(x) != x:
                print("Error: 'x' is not an integer.")
                return np.nan
            else:
                return np.matmul(
                    self.initdist,
                    np.matmul(
                        np.linalg.matrix_power(self.phgen, (x - 1)), self.exitrates
                    ),
                ).item()
        else:
            return np.matmul(
                self.initdist, np.matmul(expm(self.phgen * x), self.exitrates)
            ).item()

    def __computecumprob(self, x: float) -> float:
        """
        Returns the distribution function.

        Args:
            x (float): Computes the distribution function for the scalar x.

        Returns:
            float: The computed cumulated probability.
        """

        if self.discrete:
            if int(x) != x:
                print("Error: 'x' is not an integer.")
                return np.nan
            else:
                return 1 - np.sum(
                    np.matmul(self.initdist, np.linalg.matrix_power(self.phgen, int(x)))
                )
        else:
            return 1 - np.sum(np.matmul(self.initdist, expm(self.phgen * x)))

    def __computequantile(self, p: float, tolerance: float = 1e-9) -> int | float:
        """
        The quantile function of the PH distribution.

        Args:
            p (float): The cumulated probability.
            tolerance (float, default=1e-9): The tolerance used in the numerical algorithm.

        Returns:
            int | float: The computed quantile.
        """

        if p == 1.0:
            return np.inf
        elif p < 0.0 or p > 1.0:
            return np.nan
        elif p == 0.0:
            return 0.0
        elif self.discrete:
            return self.__dphquantfun(prob=p, itermax=1000000)
        else:
            return self.__cphquantfun(prob=p, tol=tolerance, itermax=1000000)

    def __cphsample(self) -> float:
        """
        Samples a pseudo-random value from a continuous phase-type (CPH) distribution.

        Returns:
            float: The sampled value.
        """

        t = 0.0
        s = np.random.choice(self.nphases, size=1, p=self.flatinitdist)[0]
        while True:
            t += np.random.exponential(scale=1 / (-self.phgen[s, s]))
            a = self.a[s]
            s = np.random.choice((self.nphases + 1), size=1, p=a)[0]
            if s == self.nphases:
                return t

    def __dphsample(self) -> int:
        """
        Samples a pseudo-random value from a discrete phase-type (DPH) distribution.

        Returns:
            int: The sampled value.
        """

        t = 0  # Time in discrete steps
        s = np.random.choice(self.nphases, size=1, p=self.flatinitdist)[0]
        while True:
            t += 1
            a = self.a[s]
            s = np.random.choice((self.nphases + 1), size=1, p=a)[0]
            if s == self.nphases:
                return t

    def __cphquantfun(
        self, prob: float, tol: float = 1e-9, itermax: int = 1000000
    ) -> float:
        """
        Numerically evaluates the quantile function of the CPH distribution.

        Args:
            prob (float): The cumulated probability, P(X<=x).
            tol (float, default=1e-9): The tolerance for the numerical algorithm.
            itermax (int, default=1000000): The maximum number of iterations.

        Returns:
            float: The numerically evaluated quantile.
        """

        # some initial preparation
        phinv = np.linalg.inv(self.phgen)
        var = 2 * np.sum(
            np.matmul(self.initdist, np.linalg.matrix_power(phinv, 2))
        ) - np.power(np.sum(np.matmul(self.initdist, phinv)), 2)
        mean = -np.sum(np.matmul(self.initdist, phinv))

        # generate initial guess
        param1 = np.log(np.power(mean, 2) / np.sqrt(np.power(mean, 2) + var))
        param2 = np.log(1 + var / np.power(mean, 2))
        x = lognorm.ppf(prob, param2, scale=np.exp(param1))

        # improve x until convergence
        trc = 1 - np.sum(np.matmul(self.initdist, expm(self.phgen * x)))
        iter = 0
        while np.abs(trc - prob) > tol and iter < itermax:
            grad = np.matmul(
                self.initdist, np.matmul(expm(self.phgen * x), self.exitrates)
            ).item()
            x = x - (trc - prob) / grad
            trc = 1 - np.sum(np.matmul(self.initdist, expm(self.phgen * x)))
            iter += 1
        if iter == itermax:
            print(
                "Warning: Algorithm terminated with iter==itermax. Results might be misleading."
            )
        return np.round(x, int(-np.log10(tol)) - 1)

    def __dphquantfun(self, prob: float, itermax: int = 1000000) -> int:
        """
        Numerical quantile function for the DPH distribution.

        Args:
            prob (float): The cumulated probability, P(X<=x).
            itermax (int, default=1000000): The maximum number of iterations.

        Returns:
            int: The computed quantile.
        """

        # initialization
        x = int(-1)
        trc = 0.0
        iter = 0

        # improve x until convergence
        while trc < prob and iter < itermax:
            x += 1
            trc = 1 - np.sum(
                np.matmul(self.initdist, np.linalg.matrix_power(self.phgen, x))
            )
            iter += 1
        if iter == itermax:
            print(
                "Warning: Algorithm terminated with iter==itermax. Results might be misleading."
            )
        return x


import numpy as np


class fitdph:
    """
    Fits discrete-time phase-type distributions using the
    EM algorithm from p. 675 Bladt and Nielsen (2017).
    
    References:
        Bladt, M., & Nielsen, B. F. (2017). Matrix-Exponential Distributions in Applied Probability.
        Springer. https://doi.org/10.1007/978-1-4939-7049-0    
    """

    def __init__(
        self,
        obs: np.array = None,
        initpi: np.array = None,
        initphgen: np.array = None,
        initexitrates: np.array = None,
        randominit: bool = True,
        seed: int = None,
        tolerance: float = 1e-6,
        itermax: int = 1000000,
        verbose: bool = False,
    ) -> None:
        """
        Initializes the discrete-time phase-type distribution fitter.

        Args:
            obs (np.array): Observed realizations of the phase-type distribution.
            initpi (np.array): Initial distribution vector.
            initphgen (np.array): Initial phase-type transition matrix.
            initexitrates (np.array): Initial exit probability vector.
            randominit (bool, default=True): Whether to randomize initial parameters.
            seed (int): Random seed for reproducibility.
            tolerance (float, default=1e-6): Convergence tolerance for the EM algorithm.
            itermax (int, default=1000000): Maximum number of EM iterations.
            verbose (bool, default=False): Whether to print intermediate fitting information.

        Returns:
            None
        """
        self.obs = obs  # observed realizations of the PH distribution
        self.initpi = initpi
        self.initphgen = initphgen
        self.initexitrates = initexitrates
        self.nphases = self.initphgen.shape[0]
        self.randominit = randominit
        self.seed = seed
        self.itermax = itermax
        self.tolerance = tolerance
        self.verbose = verbose
        self.__initialize()

    def fit(self) -> None:
        """
        Fits the discrete-time phase-type distribution using the EM algorithm.

        Args:
            None

        Returns:
            None
        """
        # fit the DPH distribution
        iter = 0
        eps = np.inf
        loglik0 = self.loglikelihood
        while iter < self.itermax and eps > self.tolerance:
            self.__estep()
            self.__mstep()
            self.__updatelikelihood()
            eps = self.loglikelihood - loglik0
            loglik0 = self.loglikelihood
            iter += 1
            if self.verbose and iter % 25 == 0:
                if isinstance(eps, float):
                    printeps = eps
                else:
                    printeps = eps.item()
                print(
                    "iter =",
                    iter,
                    "  eps =",
                    printeps,
                    "  mean =",
                    self.getmean(),
                    "  var =",
                    self.getvar(),
                )
        self.__polish()

    def getinitdist(self) -> np.array:
        """
        Returns the initial distribution vector.

        Args:
            None

        Returns:
            np.array: The initial distribution.
        """
        return self.pi

    def getphasegen(self) -> np.array:
        """
        Returns the phase-type transition matrix.

        Args:
            None

        Returns:
            np.array: The phase-type generator matrix.
        """
        return self.phgen

    def getexitrates(self) -> np.array:
        """
        Returns the exit probability vector.

        Args:
            None

        Returns:
            np.array: The exit rates.
        """
        return self.exitrates

    def getmean(self) -> float:
        """
        Returns the mean of the discrete-time phase-type distribution.

        Args:
            None

        Returns:
            float: The mean of the distribution.
        """
        return np.sum(
            np.matmul(
                self.pi, np.linalg.inv(np.subtract(np.eye(self.nphases), self.phgen))
            )
        )

    def getvar(self) -> float:
        """
        Returns the variance of the discrete-time phase-type distribution.

        Args:
            None

        Returns:
            float: The variance of the distribution.
        """
        Tinv = np.linalg.inv(np.subtract(np.eye(self.nphases), self.phgen))
        return (
            np.sum(
                np.matmul(
                    np.matmul(self.pi, Tinv),
                    np.subtract((2 * Tinv), np.eye(self.nphases)),
                )
            )
            - self.getmean() ** 2
        )

    def getdensity(self, x: int) -> float:
        """
        Returns the probability mass function evaluated at x.

        Args:
            x (int): The discrete time point at which the density is evaluated.

        Returns:
            float: The computed probability mass.
        """
        if int(x) != x:
            print("Error: 'x' is not an integer.")
            return np.nan
        else:
            return np.matmul(
                self.pi,
                np.matmul(np.linalg.matrix_power(self.phgen, (x - 1)), self.exitrates),
            ).item()

    def getcumprob(self, x: int) -> float:
        """
        Returns the cumulative distribution function P(X ≤ x).

        Args:
            x (int): The discrete time point at which the cumulative probability is evaluated.

        Returns:
            float: The cumulative probability.
        """
        if int(x) != x:
            print("Error: 'x' is not an integer.")
            return np.nan
        else:
            return 1 - np.sum(np.matmul(self.pi, np.linalg.matrix_power(self.phgen, x)))

    def getloglik(self) -> float:
        """
        Returns the log-likelihood of the fitted model.

        Args:
            None

        Returns:
            float: The log-likelihood value.
        """
        return self.loglikelihood

    def getaic(self) -> float:
        """
        Returns Akaike's Information Criterion (AIC).

        Args:
            None

        Returns:
            float: The AIC value.
        """
        return -2 * self.loglikelihood + 2 * self.nparam

    def getbic(self) -> float:
        """
        Returns the Bayesian Information Criterion (BIC).

        Args:
            None

        Returns:
            float: The BIC value.
        """
        return -2 * self.loglikelihood + self.nparam * np.log(len(self.obs))

    def __initialize(self) -> None:
        """
        Initializes internal parameters and prepares the model for fitting.

        Args:
            None

        Returns:
            None
        """
        if self.seed is not None:
            np.random.seed(self.seed)

        self.obs = self.obs.astype(int)
        self.obs = np.sort(self.obs)
        self.initpi = self.initpi.astype(float)
        self.initphgen = self.initphgen.astype(float)
        self.initexitrates = self.initexitrates.astype(float)
        self.identity = np.eye(self.nphases)

        self.pi = np.array(self.initpi, dtype=float).flatten()
        self.phgen = self.initphgen
        self.exitrates = np.array(self.initexitrates, dtype=float).flatten()

        if self.randominit:
            self.__initrandom()
        self.__updatelikelihood()
        self.__countParameters()

    def __initrandom(self) -> None:
        """
        Randomly initializes the parameters of the discrete-time phase-type distribution.

        Args:
            None

        Returns:
            None
        """
        nzidx = np.nonzero(self.pi)
        u = np.random.uniform(low=0.0, high=1.0, size=len(nzidx))
        u = u / np.sum(u)
        self.pi[nzidx] = u

        nzidx = np.nonzero(self.exitrates)
        u = np.random.uniform(low=0.0, high=1.0, size=len(nzidx))
        self.exitrates[nzidx] = u

        for i in range(self.nphases):
            nzidx = np.nonzero(np.ravel(self.phgen[i, :]))
            u = np.random.uniform(low=0.0, high=1.0, size=nzidx[0].size)
            u = (u / np.sum(u)) * (1.0 - self.exitrates[i])
            self.phgen[i, nzidx] = u

    def __estep(self) -> None:
        """
        Performs the expectation (E) step of the EM algorithm.

        Args:
            None

        Returns:
            None
        """
        self.bi = np.zeros(self.nphases)
        self.ni = np.zeros(self.nphases)
        self.nij = np.zeros((self.nphases, self.nphases))
        self.phgeninv = np.linalg.inv(self.phgen)

        y0 = -1
        for y in self.obs:
            if y != y0:
                if y == 1:
                    Tpow = self.identity
                    Ttprod = self.exitrates.flatten()
                else:
                    Tpow = np.linalg.matrix_power(self.phgen, y - 1)
                    Ttprod = np.ravel(np.matmul(Tpow, self.exitrates))

                piTtprod = np.matmul(self.pi, Ttprod)

                if y >= 2:
                    self.__Kmatrix(y)

            y0 = y

            if piTtprod != 0.0:
                piTpow = np.ravel(np.matmul(self.pi, Tpow))

                for i in range(self.nphases):
                    self.bi[i] += self.pi[i] * Ttprod[i] / piTtprod
                    self.ni[i] += piTpow[i] * self.exitrates[i] / piTtprod

                    if y >= 2:
                        for j in range(self.nphases):
                            self.nij[i, j] += (self.phgen[i, j] * self.Kmat[j, i]) / piTtprod
                            
    def __mstep(self) -> None:
        """
        Performs the maximization (M) step of the EM algorithm.

        Args:
            None

        Returns:
            None
        """
        for i in range(self.nphases):
            self.pi[i] = self.bi[i] / self.obs.size

        for i in range(self.nphases):
            sm = self.ni[i]
            for j in range(self.nphases):
                sm += self.nij[i, j]
            self.exitrates[i] = self.ni[i] / sm

        for i in range(self.nphases):
            for j in range(self.nphases):
                sm = self.ni[i]
                for k in range(self.nphases):
                    sm += self.nij[i, k]
                self.phgen[i, j] = self.nij[i, j] / sm

    def __updatelikelihood(self) -> None:
        """
        Updates the log-likelihood based on current model parameters.

        Args:
            None

        Returns:
            None
        """
        self.loglikelihood = 0.0
        for y in self.obs:
            self.loglikelihood += np.log(self.__getProbMass(y)).item()

    def __getProbMass(self, y: int) -> float:
        """
        Computes the probability mass at observation y.

        Args:
            y (int): Discrete observation value.

        Returns:
            float: The probability mass.
        """
        t = self.exitrates[:, None]
        return np.matmul(
            self.pi, np.matmul(np.linalg.matrix_power(self.phgen, y), t)
        )

    def __Kmatrix(self, y: int) -> None:
        """
        Computes the matrix-function K for a given observation.

        Args:
            y (int): Observation value.

        Returns:
            None
        
        if y < 2:
            return

        self.Kmat = np.zeros((self.nphases, self.nphases))

        Ty = np.linalg.matrix_power(self.phgen, y - 2)
        exit_prod = np.matmul(Ty, self.exitrates)
        pi_mat = self.pi

        for k in range(y - 1):
            self.Kmat += np.outer(exit_prod, pi_mat)
            if k < y - 2:
                Ty = np.matmul(Ty, self.phgeninv)
                exit_prod = np.matmul(Ty, self.exitrates)
                pi_mat = np.matmul(pi_mat, self.phgen)
        """

        if y < 2:
            self.Kmat = np.zeros((self.nphases, self.nphases))
            return

        self.Kmat = np.zeros((self.nphases, self.nphases))

        Ty = np.linalg.matrix_power(self.phgen, y - 2)

        exit_prod = np.matmul(Ty, self.exitrates).flatten()
        pi_mat = self.pi.flatten()

        for k in range(y - 1):
            self.Kmat += np.outer(exit_prod, pi_mat)

            if k < y - 2:
                Ty = np.matmul(Ty, self.phgeninv)
                exit_prod = np.matmul(Ty, self.exitrates).flatten()
                pi_mat = np.matmul(pi_mat, self.phgen).flatten()        


    def __countParameters(self) -> None:
        """
        Counts the number of independent model parameters.

        Args:
            None

        Returns:
            None
        """
        phg = 0
        for i in range(self.nphases):
            phg += (
                np.count_nonzero(self.phgen[i, :])
                + np.count_nonzero(self.exitrates[i])
                - 1
            )
        self.nparam = phg + (np.count_nonzero(self.pi) - 1)

    def __polish(self) -> None:
        """
        Normalizes model parameters.

        Args:
            None

        Returns:
            None
        """
        self.pi = self.pi / np.sum(self.pi)



import sys
import numpy as np
import matplotlib.pyplot as plt
from statsmodels.stats.proportion import proportion_confint
from phasedist.fitcph import fitcph
from phasedist.fitdph import fitdph
from phasedist.dist import dist


class fit:
    """
    Fit continuous or discrete-time phase-type (PH) distributions to data.

    This class provides a unified interface for estimating the parameters of
    continuous (CPH) or discrete (DPH) phase-type distributions using the
    Expectation-Maximization (EM) algorithm. Users may specify initial
    parameters, impose specific structural constraints (e.g., Coxian,
    hyper-exponential), or rely on random initialization.

    The fitted distribution is returned as a `phasedist.dist.dist`
    object that supports evaluation of densities, CDFs, quantiles, and random
    sampling.
    """
    
    def __init__(
        self,
        obs: np.array = None,
        nphases: int = 2,
        dtype: str = "general",
        discrete: bool = False,
        initdist: np.array = None,
        initphgen: np.array = None,
        initexitrates: np.array = None,
        randominit: bool = True,
        seed: int = None,
        tolerance: float = 1e-6,
        itermax: int = 1000000,
        verbose: bool = False,
    ) -> None:
        """
        Initialize the fitting procedure for a phase-type distribution.

        Args:
            obs (np.array):
                Array of observed data points.
            nphases (int, default=2):
                Number of phases in the PH distribution.
            dtype (str, default="general"):
                Distribution type (e.g., "general", "hyperexp").
            discrete (bool, default=False):
                Whether a discrete PH distribution is assumed.
            initdist (np.array, optional):
                Initial distribution vector for custom structure or initialization.
            initphgen (np.array, optional):
                Initial phase-type generator matrix for custom structure or initialization.
            initexitrates (np.array, optional):
                Initial exit rate vector for custom structure or initialization.
            randominit (bool, default=True):
                Whether to use random initialization instead of provided initial parameters.
            seed (int, optional):
                Random seed.
            tolerance (float, default=1e-6):
                Convergence tolerance for the EM algorithm.
            itermax (int, default=1000000):
                Maximum number of iterations.
            verbose (bool, default=False):
                If True, prints progress output during fitting.
                
        Notes
        -----        
        Input validation is performed automatically. If validation fails,
        program execution terminates.        
        """

        # set parameters
        self.obs = obs
        self.nphases = nphases
        self.dtype = dtype
        self.discrete = discrete
        self.initdist = initdist
        self.initphgen = initphgen
        self.initexitrates = initexitrates
        self.randominit = randominit
        self.seed = seed
        self.tolerance = tolerance
        self.itermax = itermax
        self.verbose = verbose

        # checking and fitting
        if self.__checkinputs():
            self.__fitdist()  # fit the parameters
        else:
            sys.exit(1)  # terminate the program

    # ----------------------------------------------------------------------
    #   PUBLIC METHODS
    # ----------------------------------------------------------------------

    def getinitdist(self) -> np.array:
        """
        Return the fitted initial distribution vector.

        Returns:
            np.array: Row vector representing the initial phase probabilities.
        """        
        return self.dist.getinitdist()

    def getphasegen(self) -> np.array:
        """
        Return the fitted phase-type generator matrix.

        Returns:
            np.array: Sub-intensity matrix (CPH) or sub-transition matrix (DPH).
        """
        return self.dist.getphasegen()

    def getexitrates(self) -> np.array:
        """
        Return the fitted exit rate vector.

        Returns:
            np.array: Column vector of exit rates from each phase.
        """
        return self.dist.getexitrates()

    def getmean(self) -> float:
        """
        Return the mean of the fitted PH distribution.

        Returns:
            float: The mean of the distribution.
        """
        return self.dist.getmean()

    def getvar(self) -> float:
        """
        Return the variance of the fitted PH distribution.

        Returns:
            float: The variance of the distribution.
        """
        return self.dist.getvar()

    def getdensity(self, x: float) -> float:
        """
        Evaluate the fitted distribution's density at point x.

        Args:
            x (float): Evaluation point.

        Returns:
            float: Probability density (CPH) or probability mass (DPH).
        """
        return self.dist.getdensity(x)

    def getcumprob(self, x: float) -> float:
        """
        Evaluate the fitted cumulative distribution function at point x.

        Args:
            x (float): Evaluation point.

        Returns:
            float: Cumulative distribution value P(X<=x).
        """
        return self.dist.getcumprob(x)

    def getquantile(self, p: float, tolerance: float = 1e-6) -> int | float:
        """
        Compute the p-quantile of the fitted PH distribution.

        Args:
            p (float): Cumulative probability in the interval (0, 1).
            tolerance (float, default=1e-6): Numerical tolerance for the quantile search.

        Returns:
            int | float: The p-quantile.
        """
        return self.dist.getquantile(p, tolerance)

    def getloglik(self) -> float:
        """
        Return the log-likelihood for the fitted PH model.

        Returns:
            float: Log-likelihood evaluated at the fitted parameters.
        """        
        return self.d.getloglik()

    def getaic(self) -> float:
        """
        Compute Akaike's Information Criterion (AIC).

        Returns:
            float: The AIC value of the fitted model.
        """
        return self.d.getaic()

    def getbic(self) -> float:
        """
        Compute Bayesian Information Criterion (BIC).

        Returns:
            float: The BIC value of the fitted model.
        """
        return self.d.getbic()

    def getdist(self) -> dist:
        """
        Return the fitted PH distribution as a dist object.

        Returns:
            dist: Fully constructed phase-type distribution object.
        """
        return self.dist

    def plot(
        self,
        confint: bool = True,
        confidence: float = 0.95,
        xlabel: str = "x",
        ylabel: str = "CDF",
        title: str = "Empirical and Fitted CDFs",
        labelfitted: str = "Fitted CDF",
        labelempirical: str = "Empirical CDF",
        filename: str = "CDFcheck.png"
    ) -> None:
        """
        Plot empirical and fitted cumulative distribution functions (CDFs).

        Args:
            confint (bool, default=True): Whether to include confidence intervals for the empirical CDF.
            confidence (float, default=0.95): Confidence level for the intervals.
            xlabel (str, default="x"): Label for the x-axis.
            ylabel (str, default="CDF"): Label for the y-axis.
            title (str, default="Empirical and Fitted CDFs"): Plot title.
            labelfitted (str, default="Fitted CDF"): Legend label for the fitted CDF.
            labelempirical (str, default="Empirical CDF"): Legend label for the empirical CDF.

        Returns:
            None: Displays the plot.
        """

        obssorted = np.sort(self.obs)

        if not self.discrete:
            empcdf = np.arange(1, self.obs.size + 1)
            empcdf_upper = np.zeros(empcdf.size)
            empcdf_lower = np.zeros(empcdf.size)
            for i in range(empcdf.size):
                lims = proportion_confint(
                    empcdf[i], self.obs.size, alpha=(1 - confidence), method="wilson"
                )
                empcdf_upper[i] = lims[1]
                empcdf_lower[i] = lims[0]
            empcdf = empcdf.astype(float)
            empcdf /= float(self.obs.size)
            res = 1000
            theocdf = np.zeros(res)
            x = np.linspace(np.min(obssorted), np.max(obssorted), res)
            for i in range(res):
                theocdf[i] = self.getcumprob(x[i])
        else:
            x = np.arange(np.min(obssorted), np.max(obssorted) + 1)
            empcdf = np.zeros(x.size)
            empcdf_upper = np.zeros(x.size)
            empcdf_lower = np.zeros(x.size)
            theocdf = np.zeros(x.size)
            emp_old = 0
            for i in range(x.size):
                theocdf[i] = self.getcumprob(x[i])
                a = obssorted[obssorted == (i + np.min(obssorted))]
                empcdf[i] = a.size + emp_old
                emp_old = empcdf[i]
                lims = proportion_confint(
                    empcdf[i],
                    self.obs.size,
                    alpha=(1 - confidence),
                    method="binom_test",
                )
                empcdf_upper[i] = lims[1]
                empcdf_lower[i] = lims[0]
            empcdf = empcdf.astype(float)
            empcdf /= float(self.obs.size)

        plt.figure(figsize=(8, 6))

        if not self.discrete:
            plt.plot(x, theocdf, label=labelfitted, lw=1, linestyle="-", color="blue")
            plt.plot(
                obssorted,
                empcdf,
                label=labelempirical,
                lw=1,
                linestyle="-",
                color="red",
            )
            if confint:
                plt.plot(
                    obssorted,
                    empcdf_upper,
                    label="Upper conf. int.",
                    lw=1,
                    linestyle="--",
                    color="red",
                )
                plt.plot(
                    obssorted,
                    empcdf_lower,
                    label="Lower conf. int.",
                    lw=1,
                    linestyle="--",
                    color="red",
                )
        else:
            plt.scatter(x, empcdf, label=labelempirical, lw=1, marker="x", color="red")
            if confint:
                plt.scatter(
                    x,
                    empcdf_upper,
                    label="Upper conf. int.",
                    marker="_",
                    s=100,
                    color="red",
                )
                plt.scatter(
                    x,
                    empcdf_lower,
                    label="Lower conf. int.",
                    marker="_",
                    s=100,
                    color="red",
                )
            plt.scatter(x, theocdf, label=labelfitted, lw=1, marker="x", color="blue")

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(title)
        plt.legend()
        plt.grid()
        plt.savefig(filename)

    # ----------------------------------------------------------------------
    #   PRIVATE METHODS
    # ----------------------------------------------------------------------

    def __checkinputs(self) -> bool:
        """
        Validate input parameters prior to fitting.

        Returns:
            bool: True if all inputs are valid; otherwise False.
        """

        # check data types and convert if necesarry
        if isinstance(self.obs, list):
            self.obs = np.array(self.obs)
        elif not isinstance(self.obs, np.ndarray):
            print(
                "Error: Observations can only be specified as a list or a NumPy array."
            )
            return False
        if not isinstance(self.nphases, int) or self.nphases < 1:
            print(
                "Error: The number of phases can only be specified as an integer larger than 0."
            )
            return False
        if not isinstance(self.dtype, str):
            print("Error: The distribution type can only be specified as a string.")
        if not isinstance(self.discrete, bool):
            print("Error: The argument 'discrete' needs to be of type 'bool'.")
        if self.initdist is not None and (
            isinstance(self.initdist, np.ndarray) or isinstance(self.initdist, list)
        ):
            self.initdist = np.matrix(self.initdist)
        elif self.initdist is not None and not isinstance(self.initdist, np.matrix):
            print(
                "Error: The initial distribution can only be specified as a list, NumPy array, or a NumPy matrix."
            )
            return False
        if self.initphgen is not None and (
            isinstance(self.initphgen, np.ndarray) or isinstance(self.initphgen, list)
        ):
            self.initphgen = np.matrix(self.initphgen)
        elif self.initphgen is not None and not isinstance(self.initphgen, np.matrix):
            print(
                "Error: The PH generator can only be specified as a list or a NumPy matrix."
            )
            return False
        if self.initexitrates is not None and (
            isinstance(self.initexitrates, np.ndarray)
            or isinstance(self.initexitrates, list)
        ):
            self.initexitrates = np.transpose(np.matrix(self.initexitrates))
        elif self.initexitrates is not None and not isinstance(
            self.initexitrates, np.matrix
        ):
            print(
                "Error: The exit rate vector can only be specified as a list, NumPy array, or a NumPy matrix."
            )
            return False
        if not isinstance(self.randominit, bool):
            print("Error: The argument 'randominit' needs to be of type 'bool'.")
            return False
        if self.seed is not None and not isinstance(self.seed, int):
            print("Error: The seed can only be specified as an integer.")
            return False
        if not isinstance(self.tolerance, float):
            print("Error: The argument 'tolerance' needs to be of type 'float'.")
            return False
        if not isinstance(self.itermax, float) and not isinstance(self.itermax, int):
            print("Error: The argument 'itermax' needs to be of type 'float' or 'int'.")
            return False
        if not isinstance(self.verbose, bool):
            print("Error: The argument 'verbose' needs to be of type 'bool'.")
            return False
        if self.discrete and self.dtype == "general":
            print(
                "Error: The 'general' option is currently inactivated for discrete PH distributions."
            )
            return False

        # check observations are equal to or greather than zero
        if np.any(self.obs < 0):
            print("Error: The array of observations contains negative values.")
            return False

        # check PH generator and exit rates in case of no random initialization
        if self.dtype == "custom" or not self.randominit:
            self.nphases = self.initphgen.shape[0]
        if not self.randominit:
            if not self.__correctphgen(
                self.initphgen, self.initexitrates
            ) or not self.__correctinitdist(self.initdist):
                return False

        return True

    def __fitdist(self) -> int:
        """
        Fit the PH distribution parameters using the chosen model type and EM algorithm.

        Returns:
            int: 0 if successful, 1 otherwise.
        """

        # set distribution type
        if self.dtype == "general":
            self.__general()
        elif self.dtype == "generlang":
            self.__generlang()
        elif self.dtype == "hyperexp":
            self.__hyperexp()
        elif self.dtype == "coxian":
            self.__coxian()
        elif self.dtype == "gencoxian":
            self.__gencoxian()
        elif self.dtype != "custom":
            print("Error: Unknown distribution type.")
            return 1

        # check for zeros in observations
        obsnonzero, fraczero = self.__checkzeros()

        # fit parameters
        if self.discrete:
            self.d = fitdph(
                obs=obsnonzero,
                initpi=self.initdist,
                initphgen=self.initphgen,
                initexitrates=self.initexitrates,
                randominit=self.randominit,
                seed=self.seed,
                tolerance=self.tolerance,
                itermax=self.itermax,
                verbose=self.verbose,
            )
        else:

            self.d = fitcph(
                obs=obsnonzero,
                initpi=self.initdist,
                initphgen=self.initphgen,
                initexitrates=self.initexitrates,
                randominit=self.randominit,
                seed=self.seed,
                tolerance=self.tolerance,
                itermax=self.itermax,
                verbose=self.verbose,
            )
        self.d.fit()

        # check fitted parameters
        self.__checkfit()

        # adjust for zeros in observations
        self.d.initpi = self.d.initpi * (1 - fraczero)

        # create object for output PH distribution
        self.dist = dist(
            discrete=self.discrete,
            initdist=self.d.getinitdist(),
            phgen=self.d.getphasegen(),
            seed=self.seed,
        )

        if self.fitaccepted:
            return 0
        else:
            print(
                "The PH distribution might contain infeasible or inaccurate parameters."
            )
            return 1

    def __general(self) -> None:
        """
        Initialize parameters for a fully general PH distribution.
        
        Returns:
            None: Initialized parameters.
        """
        
        self.initdist = np.matrix(np.ones((1, self.nphases)))
        self.initphgen = np.matrix(np.ones((self.nphases, self.nphases)))
        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))

    def __generlang(self) -> None:
        """
        Initialize parameters for a generalized Erlang distribution.
        
        Returns:
            None: Initialized parameters.
        """
        
        self.initdist = np.matrix(np.zeros((1, self.nphases)))
        self.initdist[0, 0] = 1

        self.initexitrates = np.matrix(np.zeros((self.nphases, 1)))
        self.initexitrates[self.nphases - 1, 0] = 1

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1
            if i < (self.nphases - 1):
                self.initphgen[i, i + 1] = 1

    def __hyperexp(self) -> None:
        """
        Initialize parameters for a hyper-exponential distribution.
        
        Returns:
            None: Initialized parameters.
        """
        self.initdist = np.matrix(np.ones((1, self.nphases)))

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1

    def __coxian(self) -> None:
        """
        Initialize parameters for a Coxian distribution.
        
        Returns:
            None: Initialized parameters.
        """
        self.initdist = np.matrix(np.zeros((1, self.nphases)))
        self.initdist[0, 0] = 1

        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1
            if i < (self.nphases - 1):
                self.initphgen[i, i + 1] = 1

    def __gencoxian(self) -> None:
        """
        Initialize parameters for a generalized Coxian distribution.
        
        Returns:
            None: Initialized parameters.
        """
        self.initdist = np.matrix(np.ones((1, self.nphases)))
        self.initexitrates = np.matrix(np.ones((self.nphases, 1)))

        self.initphgen = np.matrix(np.zeros((self.nphases, self.nphases)))
        for i in range(self.nphases):
            self.initphgen[i, i] = 1
            if i < (self.nphases - 1):
                self.initphgen[i, i + 1] = 1

    def __checkzeros(self):
        """
        Detect zero observations and compute their empirical proportion.

        Returns:
            tuple: (nonzero observations, proportion of zeros)
        """
        nz = np.nonzero(self.obs)[0]
        return self.obs[nz], 1 - (nz.size / self.obs.size)

    def __checkfit(self) -> None:
        """
        Check feasibility of the fitted PH parameters.
        
        Returns:
            None: Feasibility check.
        """
        self.fitaccepted = True
        if self.__correctphgen(
            self.d.getphasegen(), self.d.getexitrates()
        ) and self.__correctinitdist(self.d.getinitdist()):
            self.fitaccepted = True
        else:
            self.fitaccepted = False

    def __correctphgen(self, phasegen: np.array, exitrates: np.array) -> bool:
        """
        Validate a PH generator matrix and exit rate vector.

        Args:
        phasegen (np.array): Candidate PH generator matrix.
        exitrates (np.array): Candidate exit rate vector.

        Returns:
            bool: True if the input is feasible.
        """

        # check PH generator
        if phasegen.shape[0] != phasegen.shape[1] or phasegen.shape[0] != self.nphases:
            print(
                "Error: The dimensions of the PH generator does not match the number of phases."
            )
            return False
        if (
            np.where(np.isnan(phasegen))[0].size > 0
            or np.where(np.isinf(phasegen))[0].size > 0
            or np.where(np.isneginf(phasegen))[0].size > 0
        ):
            print("Error: The PH generator contains NaN or/and infinity values.")
            return False
        if np.any(phasegen[~np.eye(self.nphases, dtype=bool)] < 0):
            print("Error: The PH generator contains negative off-diagonal values.")
            return False
        if not self.discrete and np.any(phasegen[np.eye(self.nphases, dtype=bool)] > 0):
            print("Error: The PH generator contains positive diagonal values.")
            return False
        
        # check exit rates
        if self.nphases != exitrates.size:
            print(
                "Error: The size of the exit rate vector does not match the number of phases."
            )
            return False
        if (
            np.where(np.isnan(exitrates))[0].size > 0
            or np.where(np.isinf(exitrates))[0].size > 0
            or np.where(np.isneginf(exitrates))[0].size > 0
        ):
            print("Error: The exit rate vector contains NaN or/and infinity values.")
            return False
        if np.any(exitrates < 0):
            print("Error: The exit rate vector contains negative values.")
            return False
        return True

    def __correctinitdist(self, initdist: np.array) -> bool:
        """
        Validate an initial distribution vector.

        Args:
            initdist (np.array): Candidate initial distribution.

        Returns:
            bool: True if valid.
        """

        if initdist.size != self.nphases:
            print(
                "Error: The size of the initial distribution does not match the number of phases."
            )
            return False
        if (
            np.where(np.isnan(initdist))[0].size > 0
            or np.where(np.isinf(initdist))[0].size > 0
            or np.where(np.isneginf(initdist))[0].size > 0
        ):
            print(
                "Error: The initial distribution contains NaN or/and infinity values."
            )
            return False
        if np.any(initdist < 0.0):
            print("Error: The initial distribution contains negative values.")
            return False
        if np.abs(np.sum(initdist) - 1.0) > 1e-14:
            print(
                "Warning: Prior to adjusting for zeros in the observations the initial distribution summed to "
                + str(np.sum(initdist))
            )
            return False
        return True
